@@ -216,6 +216,7 @@ struct sshfs {
 	int rename_workaround;
 	int truncate_workaround;
 	int buflimit_workaround;
+	int unrel_append;
 	int fstat_workaround;
 	int transform_symlinks;
 	int follow_symlinks;
@@ -279,7 +280,6 @@ struct sshfs {
 	int ext_statvfs;
 	int ext_hardlink;
 	int ext_fsync;
-	mode_t mnt_mode;
 	struct fuse_operations *op;
 
 	/* statistics */
@@ -410,6 +410,7 @@ static struct fuse_opt sshfs_opts[] = {
 	SSHFS_OPT("dir_cache=no",  dir_cache, 0),
 	SSHFS_OPT("writeback_cache=yes", writeback_cache, 1),
 	SSHFS_OPT("writeback_cache=no", writeback_cache, 0),
+	SSHFS_OPT("unreliable_append", unrel_append, 1),
 
 	SSHFS_OPT("-h",		show_help, 1),
 	SSHFS_OPT("--help",	show_help, 1),
@@ -423,7 +424,13 @@ static struct fuse_opt sshfs_opts[] = {
 	FUSE_OPT_KEY("-p ",            KEY_PORT),
 	FUSE_OPT_KEY("-C",             KEY_COMPRESS),
 	FUSE_OPT_KEY("-F ",            KEY_CONFIGFILE),
+
+	/* For backwards compatibility */
+	SSHFS_OPT("cache=yes", dir_cache, 1),
+	SSHFS_OPT("cache=no",  dir_cache, 0),
+	
 	FUSE_OPT_END
+
 };
 
 static struct fuse_opt workaround_opts[] = {
@@ -1644,16 +1651,11 @@ static int sftp_check_root(const char *base_path)
 	if (!(flags & SSH_FILEXFER_ATTR_PERMISSIONS))
 		goto out;
 
-	if (S_ISDIR(sshfs.mnt_mode) && !S_ISDIR(stbuf.st_mode)) {
+	if (!S_ISDIR(stbuf.st_mode)) {
 		fprintf(stderr, "%s:%s: Not a directory\n", sshfs.host,
 			remote_dir);
 		goto out;
 	}
-	if ((sshfs.mnt_mode ^ stbuf.st_mode) & S_IFMT) {
-		fprintf(stderr, "%s:%s: type of file differs from mountpoint\n",
-			sshfs.host, remote_dir);
-		goto out;
-	}
 
 	err = 0;
 
@@ -2073,11 +2075,16 @@ static int sftp_readdir_async(struct buffer *handle, void *buf, off_t offset,
 			outstanding--;
 
 			if (done) {
+				/* We need to cache want_reply, since processing
+				   thread may free req right after unlock() if
+				   want_reply == 0 */
+				int want_reply;
 				pthread_mutex_lock(&sshfs.lock);
 				if (sshfs_req_pending(req))
 					req->want_reply = 0;
+				want_reply = req->want_reply;
 				pthread_mutex_unlock(&sshfs.lock);
-				if (!req->want_reply)
+				if (!want_reply)
 					continue;
 			}
 
@@ -2490,11 +2497,29 @@ static int sshfs_open_common(const char *path, mode_t mode,
 	uint32_t pflags = 0;
 	struct iovec iov;
 	uint8_t type;
-	uint64_t wrctr;
+	uint64_t wrctr = 0;
 
 	if (sshfs.dir_cache)
 		wrctr = cache_get_write_ctr();
 
+	/* With writeback cache, kernel may send read requests even
+	   when userspace opened write-only */
+	if (sshfs.writeback_cache &&
+	    (fi->flags & O_ACCMODE) == O_WRONLY) {
+		fi->flags &= ~O_ACCMODE;
+		fi->flags |= O_RDWR;
+	}
+
+	/* Having the kernel handle O_APPEND doesn't work reliably, if
+	   the file changes on the server at the wrong time, we will
+	   overwrite data instead of appending. */
+	if ((fi->flags & O_APPEND) && sshfs.writeback_cache) {
+		if(sshfs.unrel_append)
+			fi->flags &= ~O_APPEND;
+		else
+			return -EINVAL;
+	}
+
 	if ((fi->flags & O_ACCMODE) == O_RDONLY)
 		pflags = SSH_FXF_READ;
 	else if((fi->flags & O_ACCMODE) == O_WRONLY)
@@ -2513,6 +2538,9 @@ static int sshfs_open_common(const char *path, mode_t mode,
 	if (fi->flags & O_TRUNC)
 		pflags |= SSH_FXF_TRUNC;
 
+	if (fi->flags & O_APPEND)
+		pflags |= SSH_FXF_APPEND;
+	
 	sf = g_new0(struct sshfs_file, 1);
 	list_init(&sf->write_reqs);
 	pthread_cond_init(&sf->write_finished, NULL);
@@ -3145,6 +3173,9 @@ static int sshfs_getattr(const char *path, struct stat *stbuf,
 	}		
 	if (!err) {
 		err = buf_get_attrs(&outbuf, stbuf, NULL);
+#ifdef __APPLE__
+		stbuf->st_blksize = 0;
+#endif
 		buf_free(&outbuf);
 	}
 	buf_free(&buf);
@@ -3333,6 +3364,7 @@ static void usage(const char *progname)
 "    -o sshfs_sync          synchronous writes\n"
 "    -o no_readahead        synchronous reads (no speculative readahead)\n"
 "    -o sync_readdir        synchronous readdir\n"
+"    -o unreliable_append   Enable (unreliable) O_APPEND support\n"
 "    -d, --debug            print some debugging information (implies -f)\n"
 "    -o writeback_cache=BOOL enable writeback cache {yes,no} (default: yes)\n"
 "    -o dir_cache=BOOL      enable caching of directory contents (names,\n"
@@ -3820,9 +3852,6 @@ int main(int argc, char *argv[])
 	const char *sftp_server;
 	struct fuse *fuse;
 	struct fuse_session *se;
-#if !defined(__CYGWIN__)
-	struct stat st;
-#endif
 
 #ifdef __APPLE__
 	if (!realpath(*exec_path, sshfs_program_path)) {
@@ -3848,6 +3877,7 @@ int main(int argc, char *argv[])
 	sshfs.ptyfd = -1;
 	sshfs.dir_cache = 1;
 	sshfs.writeback_cache = 1;
+	sshfs.unrel_append = 0;
 	sshfs.show_help = 0;
 	sshfs.show_version = 0;
 	sshfs.singlethread = 0;
@@ -3895,7 +3925,7 @@ int main(int argc, char *argv[])
 		fprintf(stderr, "see `%s -h' for usage\n", argv[0]);
 		exit(1);
 	}
-	
+
 	if (sshfs.idmap == IDMAP_USER)
 		sshfs.detect_uid = 1;
 	else if (sshfs.idmap == IDMAP_FILE) {
@@ -3984,18 +4014,6 @@ int main(int argc, char *argv[])
 	g_free(tmp);
 	g_free(fsname);
 
-
-#if !defined(__CYGWIN__)
-	res = stat(sshfs.mountpoint, &st);
-	if (res == -1) {
-		perror(sshfs.mountpoint);
-		exit(1);
-	}
-	sshfs.mnt_mode = st.st_mode;
-#elif defined(__CYGWIN__)
-	sshfs.mnt_mode = S_IFDIR | 0755;
-#endif
-
 	if(sshfs.dir_cache)
 		sshfs.op = cache_wrap(&sshfs_oper);
 	else
